/* Version information
 *
 * $Id: buildingblocks.cpp 27 2007-06-11 09:04:27Z thlenser $
 *
 */

#include "buildingblocks.h"


namespace sbmlevolver
{
	
#define NEXT_INPUT while((inputFile.peek() == '#') || (inputFile.peek() == '\n')) inputFile.ignore(10000, '\n');
//#define IGNORE_NEWLINE while(inputFile.peek() == '\n') inputFile.ignore(10000, '\n');
#define IGNORE_WHITESPACE while(inputFile.peek() == ' ') inputFile.ignore();

BuildingBlocks::BuildingBlocks(){
	minNumSpecies = 0;
	maxNumSpecies = 0;
	minNumReactions = 0;
	maxNumReactions = 0;
	numStoredReactions = 0;
}


BuildingBlocks::BuildingBlocks(string BBstring)
{		
	minNumSpecies = 0;
	maxNumSpecies = 0;
	minNumReactions = 0;
	maxNumReactions = 0;
	numStoredReactions = 0;

	index_transport_reactions_exists = false;
	index_non_transport_reactions_exists = false;

	string formula;
	string paraId, lowerBound, upperBound;

	string tmp_str;
	int numPara;

	istringstream inputFile;
	inputFile.str(BBstring);
	
	Parameter* para = NULL;
	Reaction* new_reaction;
	KineticLaw* kl;
	
	NEXT_INPUT
	
	// Get min/maximal number of species
	if ((inputFile >> minNumSpecies).fail()) THROW_ERROR("Minimal number of species in "+BBstring+" not apropriate!");
	NEXT_INPUT
	if ((inputFile >> maxNumSpecies).fail()) THROW_ERROR("Maximal number of species in "+BBstring+" not apropriate!");
	NEXT_INPUT

	// Get min/maximal number of reactions
	if ((inputFile >> minNumReactions).fail()) THROW_ERROR("Minimal number of reactions in "+BBstring+" not apropriate!");
	NEXT_INPUT
	if ((inputFile >> maxNumReactions).fail()) THROW_ERROR("Maximal number of reactions in "+BBstring+" not apropriate!");
	NEXT_INPUT
	
	getline(inputFile, speciesInitLowerBound);
	getline(inputFile, speciesInitUpperBound);

	if(speciesInitLowerBound == "inf") THROW_ERROR("Lower initial concentration boundary cannot be infinity!");
	if(speciesInitUpperBound == "inf") speciesInitUpperBound = LARGEST_INITIAL_CONCENTRATION_STRING;
	
	NEXT_INPUT
		
	// Get reaction types (format: numRe, numPr, numMo, formula, numPara, [paraId, lb, ub]*, [T])
	while(! inputFile.eof()){
		NEXT_INPUT
			
		getline(inputFile, tmp_str, ' ');
		IGNORE_WHITESPACE
		
		if(tmp_str == string("Species:")){
			// Read a species name and bounds for initial concentrations
			getline(inputFile, tmp_str, ' ');
			IGNORE_WHITESPACE			
			Species* sp = new Species(tmp_str);
			
			if(inputFile.peek() == 'B'){
				// boundary condition is set to true
				sp->setBoundaryCondition(true);
				inputFile.ignore();
				IGNORE_WHITESPACE
			}
			
			inputFile >> lowerBound;
			inputFile >> upperBound;
			setAttr(sp,"lower_bound",lowerBound);
			setAttr(sp,"upper_bound",upperBound);
			lowerBound.clear();
			upperBound.clear();
			species.push_back(sp);
			sp = NULL;
		}
		else{
			// Read reactions of the format:
			// Reactant0 + Blah | Modifier1 + MyModifier -> Product0 + MyProduct : KineticLaw numPara [paraId, lb, ub]* [T]
			Reaction* re = new Reaction();
			if(tmp_str == string("|")) THROW_ERROR("No reactant given, write '0' in this case!");
			while((tmp_str != string("|")) && (tmp_str != string("->")) && (tmp_str != string("<->"))){
				if(tmp_str == string("+")){
					getline(inputFile, tmp_str, ' ');
					IGNORE_WHITESPACE
				}
				
				if(tmp_str != string("0")){
					SpeciesReference* sr = re->createReactant();
					sr->setSpecies(tmp_str);
					sr->initDefaults();
				}
				
				getline(inputFile, tmp_str, ' ');
				IGNORE_WHITESPACE
			}

			if((tmp_str != string("->")) && (tmp_str != string("<->"))){ // there is a modifier
				getline(inputFile, tmp_str, ' ');
				IGNORE_WHITESPACE
						
				while((tmp_str != string("->")) && (tmp_str != string("<->"))){
					if(tmp_str == string("+")){
						getline(inputFile, tmp_str, ' ');
						IGNORE_WHITESPACE
					}
					
					if(tmp_str != string("0")){
						ModifierSpeciesReference* msr = re->createModifier();
						msr->setSpecies(tmp_str);
					}
					
					getline(inputFile, tmp_str, ' ');
					IGNORE_WHITESPACE
				}
			}

			// check if reaction is reversible
			if(tmp_str == string("<->")){
				re->setReversible(true);
			}
			else if(tmp_str == string("->")){
				re->setReversible(false);
			}
			else{
				THROW_ERROR("Expected -> at this position"); 
			}
		
			getline(inputFile, tmp_str, ' ');
			IGNORE_WHITESPACE
		
			while(tmp_str != string(":")){
				if(tmp_str == string("+")){
					getline(inputFile, tmp_str, ' ');
					IGNORE_WHITESPACE
				}
				
				if(tmp_str != string("0")){
					SpeciesReference* sr = re->createProduct();
					sr->setSpecies(tmp_str);
					sr->initDefaults();
				}
				
				getline(inputFile, tmp_str, ' ');
				IGNORE_WHITESPACE
			}

			IGNORE_WHITESPACE
			getline(inputFile, formula, ' ');
			// Check that there is more on the line
			if(formula.find('\n') != string::npos){
				cerr << "Parameter information needed after kinetic law, at least \'0\'!" << endl;;
				exit(-1);
			}

			kl = re->createKineticLaw();
			kl->setFormula(formula);
			inputFile >> numPara;

			for(int i=0; i < numPara; i++){
				inputFile >> paraId;
				inputFile >> lowerBound;
				inputFile >> upperBound;
				if(lowerBound == "inf") THROW_ERROR("Lower parameter boundary cannot be infinity!");
				if(upperBound == "inf") upperBound = LARGEST_PARAMETER_STRING;
				para = re->getKineticLaw()->createParameter();
				para->setId(paraId);
			    	setAttr(para,"lower_bound",lowerBound);
				setAttr(para,"upper_bound",upperBound);
				para->unsetValue(); // parameter not specified here
				para = NULL;
				lowerBound.clear();
				upperBound.clear();
			}
			IGNORE_WHITESPACE
			if(inputFile.peek() == 'T'){
				inputFile.ignore();
				setAttr(re,"transport_reaction","true");
                                IGNORE_WHITESPACE
			}
			addReactionAsBuildingBlock(re);
			kl = NULL;
		}	

		// Now read to the end of the line
		/*while(inputFile.peek() != '\n'){
			cout << inputFile.peek() << endl;
			if(inputFile.peek() != ' ') THROW_ERROR("Too much data while reading buildingblocks file");
			inputFile.ignore();
		}*/
		IGNORE_WHITESPACE
		NEXT_INPUT
	}
	//print(&cout);
}

// Only the reactions themselves are explicitly allocated
BuildingBlocks::~BuildingBlocks()
{
	/*for(int i=0; i < maxNumSpecies; i++){
		delete species[i];
	}*/
	for(int i=0; i < reactions.size(); i++){
		for(int j=0; j < reactions[i].size(); j++){
			for(int k=0; k < reactions[i][j].size(); k++){
				for(int m=0; m < reactions[i][j][k].size(); m++){
					delete reactions[i][j][k][m];
				}
			}
		}
	}
}


/**
 * This function reads reaction-types from a model and stores them in the
 * BuildingBlocks object. If the "generalised" formula of a reaction in 
 * the model is not equal to an already stored formula, it is added.
 */
void BuildingBlocks::learnReactionsFromModel(Model* m){
	// Extract the species from the model
	for(int i=0; i < m->getNumSpecies(); i++){
		// Check if species is already in list
		bool found = false;
		for(int j=0; j < species.size(); j++){
			if(m->getSpecies(i)->getId() == species.at(j)->getId()) found = true;
		}
		if(!found){
			Species* sp = m->getSpecies(i);
			Species* new_sp = (Species*) sp->clone(); 
			double initConc;
			if(sp->isSetInitialConcentration()) initConc = sp->getInitialConcentration();
			else if(sp->isSetInitialAmount()) initConc = sp->getInitialAmount();
			else {
				THROW_ERROR("No initial quantity set!");
			}
			setAttr(sp, "lower_bound", initConc);
			setAttr(sp, "upper_bound", initConc);
			species.push_back(new_sp);
		}
	}


	for(int i=0; i < m->getNumReactions(); i++){
		// Extract generalised formula
		Reaction* re = m->getReaction(i);
		
		// Copy reaction here already to resolve stoichiometry issues
		re = (Reaction*) re->clone();
		
		for(int j=0; j < re->getKineticLaw()->getNumParameters(); j++){
			Parameter* par = re->getKineticLaw()->getParameter(j);
			if(par->isSetValue()){
				setAttr(par, "lower_bound", par->getValue());
				setAttr(par, "upper_bound", par->getValue());
			}
			else{
				THROW_ERROR("Found parameter with no value");
			}
		}
		
		// now remove specific molecules		
		int numRe = re->getNumReactants();
		int numPr = re->getNumProducts();
		int numMo = re->getNumModifiers();
		int index;
		
		for(int j=0; j < numRe; j++){
			string reac("Reactant");
			ostringstream converter;
			converter << j;
			reac.append(converter.str());
			converter.flush();
			
			string pattern = re->getReactant(j)->getSpecies();
			replaceSpeciesInReaction(re, pattern, reac);
		}
		
		for(int j=0; j < numPr; j++){
			string prod("Product");
			ostringstream converter;
			converter << j;
			prod.append(converter.str());
			converter.flush();
			
			string pattern = re->getProduct(j)->getSpecies();
			replaceSpeciesInReaction(re, pattern, prod);
		}
		
		for(int j=0; j < numMo; j++){
			string mod("Modifier");
			ostringstream converter;
			converter << j;
			mod.append(converter.str());
			converter.flush();
			
			string pattern = re->getModifier(j)->getSpecies();
			replaceSpeciesInReaction(re, pattern, mod);
		}

		// check if reaction transports mass between compartments
		string compartment;
		if(numRe > 0){
			compartment = m->getSpecies(re->getReactant(0)->getSpecies())->getCompartment();
		}
		else if(numPr > 0){
			compartment = m->getSpecies(re->getProduct(0)->getSpecies())->getCompartment();
		}
		else{
			THROW_ERROR("Reaction has no reactant or product!");
		}
		for(int j=0; j < numRe; j++){
			if(m->getSpecies(re->getReactant(j)->getSpecies())->getCompartment() != compartment){
				setAttr(re,"transport_reaction","true");
			}
		}
		for(int j=0; j < numPr; j++){
			if(m->getSpecies(re->getProduct(j)->getSpecies())->getCompartment() != compartment){
				setAttr(re,"transport_reaction","true");
			}
		}
		
		// check if reaction is known
		bool found = false;
		for(int i=0; i < getNumReactions(numRe, numPr, numMo); i++){
			if(equal(re, reactions.at(numRe).at(numPr).at(numMo).at(i))) found = true;
		}
		if(!found){
			addReactionAsBuildingBlock(re);
		}
		else{
			delete re;
		}
	}	
	
	
	if(species.size() > maxNumSpecies){
		THROW_WARN("More species have been loaded than are maximally allowed. I increased the number of allowed species!");
		maxNumSpecies = species.size();
	}
	print(&cout);
}



int BuildingBlocks::getNumReactions(){
	return numStoredReactions;
}

int BuildingBlocks::getReactionType(Reaction* re){
	int type = -1;
	int numRe = re->getNumReactants();
	int numPr = re->getNumProducts();
	int numMo = re->getNumModifiers();
	
	if(getNumReactions(numRe,numPr,numMo) < 1){
		//No type available for given reaction
		return -1; 
	} 
	for(int i=0; i < getNumReactions(numRe,numPr,numMo); i++){
		Reaction* re2 = (Reaction*) (getReaction(numRe,numPr,numMo,i))->clone();
		for(int j=0; j < numRe; j++){
			//cout << "Sp " << re2->getReactant(j)->getSpecies() << " " << re->getReactant(j)->getSpecies() << endl;
			replaceSpeciesInReaction(re2,re2->getReactant(j)->getSpecies(),re->getReactant(j)->getSpecies());
		}
		for(int j=0; j < numPr; j++){
			replaceSpeciesInReaction(re2,re2->getProduct(j)->getSpecies(),re->getProduct(j)->getSpecies());
		}
		for(int j=0; j < numMo; j++){
			replaceSpeciesInReaction(re2,re2->getModifier(j)->getSpecies(),re->getModifier(j)->getSpecies());
		}
		if(equal(re, re2)){
			type = getTypeName(numRe,numPr,numMo,i);
		}
		//cout << "Re2 " << re2->getKineticLaw()->getFormula() << endl;
		delete re2;
	}
	
	if(type == -1){
		//No type found for given reaction
		return -1;
	}
	if((type < -1) || (type >= getNumReactions())) THROW_ERROR("Invalid reaction type!");
	return type;
}

int BuildingBlocks::getTypeName(int numRe, int numPr, int numMo, int num){
	int counter = 0;
	for(int i=0; i < reactions.size(); i++){
		if(i > numRe) break;
		for(int j=0; j < reactions.at(i).size(); j++){
			if((i == numRe) && (j > numPr)) break;
			for(int k=0; k < reactions.at(i).at(j).size(); k++){
				if((i == numRe) && (j == numPr) && (k > numMo)) break;
				for(int m=0; m < reactions.at(i).at(j).at(k).size(); m++){
					if((i == numRe) && (j == numPr) && (k == numMo) && (m >= num)) break;
					counter++;
				}
			}
		}
	}
	return counter;	
}



int BuildingBlocks::getMinNumSpecies(){
	return minNumSpecies;
}

int BuildingBlocks::getMaxNumSpecies(){
	return maxNumSpecies;
}

int BuildingBlocks::getMinNumReactions(){
	return minNumReactions;
}

int BuildingBlocks::getMaxNumReactions(){
	return maxNumReactions;
}

string BuildingBlocks::getLowestInitialConcentration(){
	return speciesInitLowerBound;
}
	
string BuildingBlocks::getHighestInitialConcentration(){
	return speciesInitUpperBound;
}

Species* BuildingBlocks::getSpecies(int num){
	Species* sp;

	if(num >= ((int)species.size())){
		sp = new Species(getSpeciesId(num));
		setAttr(sp, "lower_bound", speciesInitLowerBound);
		setAttr(sp, "upper_bound", speciesInitUpperBound);
	}
	else{
		sp = (Species*) (species.at(num))->clone();
	}
	
	return sp;
}

Reaction* BuildingBlocks::getReaction(int re, int pr, int mo, int num){
	try{ return reactions.at(re).at(pr).at(mo).at(num);}
	catch(...){
		THROW_WARN("Warning: Out-of-range access to reactions in BuildingBlocks!");
		return NULL;
	}
}

Reaction* BuildingBlocks::getReaction(int reac_num){
	try{
		int counter = 0;
		for(int i=0; i < reactions.size(); i++){
			for(int j=0; j < reactions.at(i).size(); j++){
				for(int k=0; k < reactions.at(i).at(j).size(); k++){
					for(int m=0; m < reactions.at(i).at(j).at(k).size(); m++){
						if(reac_num == counter) return reactions.at(i).at(j).at(k).at(m);
						counter++;
					}
				}
			}
		}
	}
	catch(...){
		THROW_WARN("Warning: Out-of-range access to reactions in BuildingBlocks!");
		return NULL;
	}
	THROW_WARN("Warning: Out-of-range access to reactions in BuildingBlocks!");
	return NULL;
}

string BuildingBlocks::getSpeciesId(int num){
	ostringstream converter;
	converter << "X" << num;

	return converter.str();
}

string BuildingBlocks::getReactionId(int num){
	ostringstream converter;
	converter << "R" << num;

	return converter.str();
}

Species* BuildingBlocks::getSuitableSpecies(Model* m){
	// Check if a new species id is available in the building blocks
	if(m->getNumSpecies() >= getMaxNumSpecies()){
		return NULL;
	}
	
	Species* candidate;
	bool found = true;
	for(int i=0; i < getMaxNumSpecies(); i++){
		candidate = getSpecies(i);
		for(int j=0; j < m->getNumSpecies(); j++){
			if(candidate->getId() == m->getSpecies(j)->getId()) found = false;
		}
		if(found) return candidate;
		else{
			delete candidate;
			found = true;
		}
	}
	return NULL;
}


string BuildingBlocks::getSuitableReactionId(Model* m){
	// Check if a new reaction id is available
	if(m->getNumReactions() >= getMaxNumReactions()){
		return string();
	}
	
	string candidate;
	bool found = true;
	for(int i=0; i < getMaxNumReactions(); i++){
		candidate = getReactionId(i);
		for(int j=0; j < m->getNumReactions(); j++){
			if((candidate == m->getReaction(j)->getId()) || ((candidate + string("_f")) == m->getReaction(j)->getId()) || ((candidate + string("_b")) == m->getReaction(j)->getId())) found = false;
		}
		if(found) return candidate;
		else found = true;
	}
	return string();
}


int BuildingBlocks::getNumReactions(int re, int pr, int mo){
	try{ return reactions.at(re).at(pr).at(mo).size();}
	catch(...){
		return 0;
	}
}

KineticLaw* BuildingBlocks::getKineticLaw(int re, int pr, int mo, int num){
	try{ return reactions.at(re).at(pr).at(mo).at(num)->getKineticLaw();}
	catch(...){
		THROW_WARN("Warning: Out-of-range access to reactions in BuildingBlocks!");
		return NULL;
	}
}

bool BuildingBlocks::isReversible(int re, int pr, int mo, int num){
	try{ return reactions.at(re).at(pr).at(mo).at(num)->getReversible();}
	catch(...){
		THROW_ERROR("Out-of-range access to reactions in BuildingBlocks!");
	}
}


int BuildingBlocks::getNumTransportReactions(){
	// check if index exists, otherwise create it
	if(!index_transport_reactions_exists){
		int counter = 0;
		for(int i=0; i < reactions.size(); i++){
			for(int j=0; j < reactions.at(i).size(); j++){
				for(int k=0; k < reactions.at(i).at(j).size(); k++){
					for(int m=0; m < reactions.at(i).at(j).at(k).size(); m++){
						if(isSetAttr(reactions.at(i).at(j).at(k).at(m), "transport_reaction")){
							transport_reactions.push_back(counter);
						}
						counter++;
					}
				}
			}
		}
		index_transport_reactions_exists = true;
	}

	return transport_reactions.size();
}
int BuildingBlocks::getNumNonTransportReactions(){
	// check if index exists, otherwise create it
	if(!index_non_transport_reactions_exists){
		int counter = 0;
		for(int i=0; i < reactions.size(); i++){
			for(int j=0; j < reactions.at(i).size(); j++){
				for(int k=0; k < reactions.at(i).at(j).size(); k++){
					for(int m=0; m < reactions.at(i).at(j).at(k).size(); m++){
						if(!isSetAttr(reactions.at(i).at(j).at(k).at(m), "transport_reaction")){
							non_transport_reactions.push_back(counter);
						}
						counter++;
					}
				}
			}
		}
		index_non_transport_reactions_exists = true;
	}

	return non_transport_reactions.size();
}
Reaction* BuildingBlocks::getTransportReaction(int num){
	// this check also creates the index if it does not exist 
	if((num < 0) || (num >= getNumTransportReactions())){
		cerr << "Index " << num << " too large or too small for transport reactions!";
		exit(-1);
	}

	return getReaction(transport_reactions.at(num));
}
Reaction* BuildingBlocks::getNonTransportReaction(int num){
	// this check also creates the index if it does not exist 
	if((num < 0) || (num >= getNumNonTransportReactions())){
		cerr << "Index " << num << " too large or too small for non-transport reactions!";
		exit(-1);
	}

	return getReaction(non_transport_reactions.at(num));
}



void BuildingBlocks::addReactionAsBuildingBlock(Reaction* re){
	try{
		int numReactants = re->getNumReactants();
		int numModifiers = re->getNumModifiers();
		int numProducts = re->getNumProducts();
	
		// Append tree structure of reactions
		vector<Reaction*> tmp1;	
		vector< vector< Reaction* > > tmp2;
		vector< vector< vector< Reaction* > > > tmp3;
		while(reactions.size() < (numReactants+1)) reactions.push_back(tmp3);
		while(reactions.at(numReactants).size() < (numProducts+1)) reactions.at(numReactants).push_back(tmp2);	
		while(reactions.at(numReactants).at(numProducts).size() < (numModifiers+1)) reactions.at(numReactants).at(numProducts).push_back(tmp1);
	
		// Put reaction into tree structure
		//new_reaction->setId(reactionTypeName);
		reactions.at(numReactants).at(numProducts).at(numModifiers).push_back(re);
		
		numStoredReactions++;
	
	} // try
	catch(...){
		THROW_ERROR("An error occured while appending the tree of kinetic laws!");
	}
}

void BuildingBlocks::addReactionTypeAsBuildingBlock(KineticLaw* kl, int numReactants, int numProducts, int numModifiers){
	try{
	
	// Append tree structure of reactions
	vector<Reaction*> tmp1;	
	vector< vector< Reaction* > > tmp2;
	vector< vector< vector< Reaction* > > > tmp3;
	while(reactions.size() < (numReactants+1)) reactions.push_back(tmp3);
	while(reactions.at(numReactants).size() < (numProducts+1)) reactions.at(numReactants).push_back(tmp2);	
	while(reactions.at(numReactants).at(numProducts).size() < (numModifiers+1)) reactions.at(numReactants).at(numProducts).push_back(tmp1);
	
	// Put reaction into tree structure
	Reaction* new_reaction = new Reaction();
	//new_reaction->setId(reactionTypeName);
	new_reaction->setKineticLaw(kl);
	delete kl;
	kl = NULL;
	
	// Add species references
	for(int i=0; i < numReactants; i++){
		stringstream  id;
		id << "Reactant" << i;
		SpeciesReference* sp = new SpeciesReference(id.str());
		sp->initDefaults();
		new_reaction->addReactant(sp);
		delete sp;
		sp = NULL;
	}
	// Add species references
	for(int i=0; i < numModifiers; i++){
		stringstream  id;
		id << "Modifier" << i;
		ModifierSpeciesReference* sp = new ModifierSpeciesReference(id.str());
		new_reaction->addModifier(sp);
		delete sp;
		sp = NULL;
	}
	// Add species references
	for(int i=0; i < numReactants; i++){
		stringstream  id;
		id << "Product" << i;
		SpeciesReference* sp = new SpeciesReference(id.str());
		sp->initDefaults();
		new_reaction->addProduct(sp);
		delete sp;
		sp = NULL;
	}
	
	reactions.at(numReactants).at(numProducts).at(numModifiers).push_back(new_reaction);
	new_reaction = NULL;
	
	numStoredReactions++;
	
	} // try
	catch(...){
		THROW_ERROR("An error occured while appending the tree of kinetic laws!");
	}
}

void BuildingBlocks::print(ostream* out){
	*out << "# Min/Max number of species\n" << minNumSpecies << "\n" << maxNumSpecies << endl;
	*out << "# Min/Max number of reactions\n" << minNumReactions << "\n" << maxNumReactions << endl;
	*out << "# Lower and upper bound for initial concentrations\n" << speciesInitLowerBound << "\n" << speciesInitUpperBound << endl;
	*out << "# Predefined species" << endl;
	Species * sp;
	for(int i=0; i < species.size(); i++){
		sp = species.at(i);
		*out << "Species: " << sp->getId() << "\n";
	}
	*out << "# Reaction types\n";
	for(int i=0; i < reactions.size(); i++){
		for(int j=0; j < reactions[i].size(); j++){
			for(int k=0; k < reactions[i][j].size(); k++){
				for(int m=0; m < reactions[i][j][k].size(); m++){
					Reaction* re = reactions[i][j][k][m];
					
					if(re->getNumReactants() == 0) *out << "0";
					else{
						*out << re->getReactant(0)->getSpecies();
						for(int t=1; t < re->getNumReactants(); t++){
							*out << " + " << re->getReactant(t)->getSpecies();
						}
					}
					if(re->getNumModifiers() > 0){
						*out << " | " << re->getModifier(0)->getSpecies();
						for(int t=1; t < re->getNumModifiers(); t++){
							*out << " + " << re->getModifier(t)->getSpecies();
						}
					}
					if(re->getReversible()) *out << " <-> ";
					else *out << " -> ";
					if(re->getNumProducts() == 0) *out << "0";
					else{
						*out << re->getProduct(0)->getSpecies();
						for(int t=1; t < re->getNumProducts(); t++){
							*out << " + " << re->getProduct(t)->getSpecies();
						}
					}
					*out << " : ";	
					*out	<< re->getKineticLaw()->getFormula() << " "
						<< re->getKineticLaw()->getNumParameters() << " ";
					for(int p=0; p < re->getKineticLaw()->getNumParameters(); p++){
						Parameter* par = re->getKineticLaw()->getParameter(p);
						*out << par->getId() << " ";
					}
					if(isSetAttr(re,"transport_reaction")){
						*out << "T";
					}
					*out << endl;
				}
			}
		}
	}
}
	
	

} // namespace sbmlevolver
