#include "integrator.h"

using namespace sbmlevolver;


integrator_result::~integrator_result(){
	ids.clear();
	timepoints.clear();
	map<string,vector<double> >::iterator it = values.begin();
	for(; it != values.end(); it++){
		it->second.clear();
	}
	values.clear();
}


Integrator::Integrator(){
}

Integrator::~Integrator(){
}


/**
 * Integrate the model and return the time course.
 * This function is helpful for nearly every fitness function.
 * Returns NULL in case of an integration error.
 */
integrator_result* Integrator::integrateSOSlib(Model* m, double time, int printstep, bool steady_state){
 	SaveSBMLModel(m,"test.xml");
	// SOSlib types
	SBMLResults_t *results = NULL;
	cvodeSettings_t *set = NULL;
		
	/* Setting SBML ODE Solver integration parameters  */
	set = CvodeSettings_create();
	CvodeSettings_setTime(set, time, printstep);
	//CvodeSettings_setErrors(set, 1e-4, 1e-9, 2000);
	if(steady_state){
		CvodeSettings_setHaltOnSteadyState (set, 1);
		CvodeSettings_setSteadyStateThreshold(set, 1e-5); 
	}
	//CvodeSettings_setMethod(set, 1, 12); // 1: Adams-Moulon, 0: BDF (stiff, default)
	//CvodeSettings_setErrors(set, 10, 10, 2000);
	// CvodeSettings_setJacobian(set, 0); // This could be used to reduce the number of failed integrations, but I'm not sure
	
	// calling the SBML ODE Solver which returns SBMLResults.
	// Model has to be level 2!!!
	results = Model_odeSolver(m, set);
	
	if(SolverError_getNum(ERROR_ERROR_TYPE) + SolverError_getNum(FATAL_ERROR_TYPE)){
		if(results) SBMLResults_free(results);
		results = NULL;
	}

	/* Clean up */
	//SolverError_dumpAndClearErrors();
	CvodeSettings_free(set);
	SolverError_clear();

	//if(results != NULL) cout << SBMLResults_getNout(results) << " ";
	
	// return object
	integrator_result* myresult;
	if(results == NULL){
		myresult = NULL;
		//PL
	}
	else{
	 	myresult = new integrator_result();
	 	for(int i=0; i < results->time->timepoints; i++){
	 		myresult->timepoints.push_back(results->time->values[i]);
	 	}
		for(int i=0; i < results->species->num_val; i++){
			myresult->ids.push_back(string(results->species->tc[i]->name));
			myresult->values.insert(make_pair(string(results->species->tc[i]->name), vector<double>(results->species->tc[i]->timepoints)));
			if((results->species->tc[i]->timepoints) != (myresult->timepoints.size())){
				cerr << results->species->tc[i]->timepoints << " != " << (myresult->timepoints.size()) << endl;
				THROW_ERROR("Unequal number of time points involved");
			}
			for(int j=0; j < results->time->timepoints; j++){
				myresult->values[results->species->tc[i]->name][j] = results->species->tc[i]->values[j];
			}
		}
		SBMLResults_free(results);
		results = NULL;
	}
	
	return myresult;
}

// it is important to return the point as it can be changed inside the function, which does not work via call-by-reference
integrator_result*  Integrator::integrateSOSlibContinue(Model* m, double time, int printstep, integrator_result* old_res){
    // SOSlib types
	SBMLResults_t *results = NULL;
	cvodeSettings_t *set = NULL;
		
	// Setting SBML ODE Solver integration parameters 
	set = CvodeSettings_create();
	CvodeSettings_setTime(set, time, printstep);
	//CvodeSettings_setErrors(set, 1e-4, 1e-9, 2000);
	//if(steady_state){
	//	CvodeSettings_setHaltOnSteadyState (set, 1);
	//	CvodeSettings_setSteadyStateThreshold(set, 1e-5); 
	//}
	//CvodeSettings_setMethod(set, 1, 12); // 1: Adams-Moulon, 0: BDF (stiff, default)
	//CvodeSettings_setErrors(set, 10, 10, 2000);
	// CvodeSettings_setJacobian(set, 0); // This could be used to reduce the number of failed integrations, but I'm not sure
	
	// calling the SBML ODE Solver which returns SBMLResults.
	// Model has to be level 2!!!
	results = Model_odeSolver(m, set);
	
	if(SolverError_getNum(ERROR_ERROR_TYPE) + SolverError_getNum(FATAL_ERROR_TYPE)){
		if(results) SBMLResults_free(results);
		results = NULL;
	}
	
	/* Clean up */
	//SolverError_dumpAndClearErrors();
	CvodeSettings_free(set);
	SolverError_clear();

	//if(results != NULL) cout << SBMLResults_getNout(results) << " ";
	
	// return object
	if(results == NULL){
		delete old_res;
		old_res = NULL;
		cout << old_res << endl;
	}
	else{
	 	for(int i=1; i < results->time->timepoints; i++){
	 		old_res->timepoints.push_back(results->time->values[i] + old_res->timepoints.at(old_res->timepoints.size()-1));
	 	}
		for(int i=0; i < results->species->num_val; i++){
			if(old_res->ids.size() != results->species->num_val) THROW_ERROR("Number of species doesn't match");
			for(int j=1; j < results->time->timepoints; j++){
				old_res->values[results->species->tc[i]->name].push_back(results->species->tc[i]->values[j]);
			}
		}
		SBMLResults_free(results);
		results = NULL;
	}
	return old_res;
}



/*integrator_result* Integrator::integrateGillespie(Model* mod, double time, int numSteps, unsigned long seed){
	if(mod->getNumReactions() == 0) return NULL; // otherwise the gillespie implementation crashes
	
	// Round all initial amounts to the nearest integer
	for(int i=0; i < mod->getNumSpecies(); i++){
		mod->getSpecies(i)->setInitialAmount(round(mod->getSpecies(i)->getInitialConcentration()));
	}
	// SaveSBMLModel(mod, "test.xml");	
	//cout << "Seed: " << seed << endl;		
	gsl_rng *r;
	//static int counter = 0;	 
  	r=gsl_rng_alloc(gsl_rng_mt19937);
  	gsl_rng_set(r,seed);
  	
  	// Prepare result object
  	// allocate all memory in place using push_back!
  	integrator_result* result = new integrator_result();
	for(int i=0; i < mod->getNumSpecies(); i++){
		result->ids.push_back(mod->getSpecies(i)->getId());
		result->values.insert(make_pair(string(mod->getSpecies(i)->getId()), vector<double>()));
	}
	  	  
  	int flag = simulate(mod,result,r,time,numSteps);
  	if(flag != 0){
  		delete result;
  		result = NULL;
  	}
  //cout << counter++ << endl;
  	gsl_rng_free(r);
  	return result;
}
*/


void Integrator::saveIntegratorResult(const char* filename, integrator_result* res){
	ofstream outfile;
	outfile.open(filename);
	
	outfile << "#t\t";
	for(int i=0; i < res->ids.size(); i++){
		outfile << res->ids.at(i) << "\t";
	}
	outfile << endl;

	for(int t=0; t < res->timepoints.size(); t++){
		outfile << res->timepoints.at(t) << "\t";
		for(int i=0; i < res->ids.size(); i++){
			outfile << res->values[res->ids.at(i)].at(t) << "\t";
		}
		outfile << endl;
	}
	
	outfile.close();
}
