/**
 * This class provides a wrapper around any integrator that is used for fitness evaluation.
 *
 * Also, the internal format for time-courses, integrator_result, is defined.
 */
 
#ifndef INTEGRATOR_H
#define INTEGRATOR_H

#include <vector> 
#include <map>
#include <fstream>

#include "SBMLTypes.h"
#include "sbmlsolver/sbmlResults.h"
#include "sbmlsolver/odeSolver.h"

#include "include.h"

#include "sbmltools.h" // For SaveSBMLModel


// For gillespie
//#include "gillespie2/simulate.h"
//#include <getopt.h>
//#include <time.h>
//#include <stdlib.h>
//#include <unistd.h>
// end for gillespie


using namespace std;
 
namespace sbmlevolver{
 
struct integrator_result{
	/// destructor
	~integrator_result();
	
	/// IDs of the species
	vector<string> ids;
	
	/// time points
	vector<double> timepoints;
	
	/// concentration values or molecule counts (treated as one for simplicity)
	map< string, vector<double> > values;
};


class Integrator{
public:
	/// Constructor
	Integrator();
	
	/// Destructor
	~Integrator();

	/// compute time course
	integrator_result* integrateSOSlib(Model* m, double time, int printstep, bool steady_state=false);
	
	/// continue integration
	integrator_result* integrateSOSlibContinue(Model* m, double time, int printstep, integrator_result* old_res);
	
	// Stochastic integration using gillespie algorithm
	//integrator_result* integrateGillespie(Model*mod, double time, int numSteps, unsigned long seed);
	
	
	/// determine whether the run should be stored
	void setStoreRun(bool storeRun);

	/// save the data points if the run has been stored
	void saveRunData(char* filename);

	/// save the data points in an integrator_result object
	void saveIntegratorResult(const char* filename, integrator_result* res);
	
private:
	
};

} // namespace

#endif // INTEGRATOR_H
