//#include "../objectivefunction.h"
// Not needed, you can use anything from ObjectiveFunction

#include "../integrator.h"

using namespace sbmlevolver;

/// Default construnctor, may use an input file
ObjectiveFunction::ObjectiveFunction(unsigned long int seed, const char* filename){
	// Initialise integrator
	integrator = new Integrator();

	nonMassConservancePenalty = 0;
	useAIC = false;
	sizePenalty = 0;
}
	
/// Construct the object from a string
ObjectiveFunction::ObjectiveFunction(string, unsigned long int seed){
	// Initialise integrator
	integrator = new Integrator();

	nonMassConservancePenalty = 0;
	useAIC = false;
	sizePenalty = 0;
}
	
/// Destructor
ObjectiveFunction::~ObjectiveFunction(){
	delete integrator;
};

/// Get the fitness of a model. Always specify this function!
double ObjectiveFunction::calculateRawFitness(Model* m){
	int fitness = 0; // count the number of wrong species

	if((m->getNumReactions() == 0) || (m->getNumSpecies() == 0)){
		return VERY_BAD_FITNESS;
	}

	// getting cvode settings 
	int printstep = 1; // 
	double time = 1000.0; // 
	integrator_result* result = NULL;
	double th = 0.000000001; // threshold between presence and absence of a species

	// Case 1: a1=1,a0=0,b1=0,b0=1,c1=1,c0=0
	// set concentrations of inflow parameters 
	m->getParameter("inflowa1")->setValue(0.0);
 	m->getParameter("inflowa0")->setValue(1.0);
 	m->getParameter("inflowb1")->setValue(1.0);
 	m->getParameter("inflowb0")->setValue(0.0);
 	m->getSpecies("c1")->setInitialConcentration(1.0);
 	m->getSpecies("c0")->setInitialConcentration(0.0);
 	m->getSpecies("d1")->setInitialConcentration(0.0);
 	m->getSpecies("d0")->setInitialConcentration(1.0);


	// run simulation
	result = integrator->integrateSOSlib(m, time, printstep, true);

	if(result == NULL){
		return VERY_BAD_FITNESS;
	}

	// evaluate
	if(result->values[string("a1")].at(1) > th) fitness++;
	if(result->values[string("a0")].at(1) < th) fitness++;
	if(result->values[string("b1")].at(1) < th) fitness++;
	if(result->values[string("b0")].at(1) > th) fitness++;
	if(result->values[string("c1")].at(1) < th) fitness++;
	if(result->values[string("c0")].at(1) > th) fitness++;
	if(result->values[string("d1")].at(1) > th) fitness++;
	if(result->values[string("d0")].at(1) < th) fitness++;

	delete result;
	result=NULL;

	// Case 2: a1=1,a0=0,b1=0,b0=1,c1=0,c0=1
	// set concentrations of inflow parameters 
	m->getParameter("inflowa1")->setValue(0.0);
 	m->getParameter("inflowa0")->setValue(1.0);
 	m->getParameter("inflowb1")->setValue(1.0);
 	m->getParameter("inflowb0")->setValue(0.0);
 	m->getSpecies("c1")->setInitialConcentration(0.0);
 	m->getSpecies("c0")->setInitialConcentration(1.0);
 	m->getSpecies("d1")->setInitialConcentration(1.0);
 	m->getSpecies("d0")->setInitialConcentration(0.0);


	// run simulation
	result = integrator->integrateSOSlib(m, time, printstep, false);

	if(result == NULL){
		return VERY_BAD_FITNESS;
	}

	// evaluate
	if(result->values[string("a1")].at(1) > th) fitness++;
	if(result->values[string("a0")].at(1) < th) fitness++;
	if(result->values[string("b1")].at(1) < th) fitness++;
	if(result->values[string("b0")].at(1) > th) fitness++;
	if(result->values[string("c1")].at(1) < th) fitness++;
	if(result->values[string("c0")].at(1) > th) fitness++;
	if(result->values[string("d1")].at(1) > th) fitness++;
	if(result->values[string("d0")].at(1) < th) fitness++;

	delete result;
	result=NULL;

	// Case 3: a1=0,a0=1,b1=0,b0=1,c1=1,c0=0
	// set concentrations of inflow parameters 
	m->getParameter("inflowa1")->setValue(1.0);
 	m->getParameter("inflowa0")->setValue(0.0);
 	m->getParameter("inflowb1")->setValue(1.0);
 	m->getParameter("inflowb0")->setValue(0.0);
 	m->getSpecies("c1")->setInitialConcentration(1.0);
 	m->getSpecies("c0")->setInitialConcentration(0.0);
 	m->getSpecies("d1")->setInitialConcentration(0.0);
 	m->getSpecies("d0")->setInitialConcentration(1.0);

	// run simulation
	result = integrator->integrateSOSlib(m, time, printstep, false);

	if(result == NULL){
		return VERY_BAD_FITNESS;
	}

	// evaluate
	if(result->values[string("a1")].at(1) < th) fitness++;
	if(result->values[string("a0")].at(1) > th) fitness++;
	if(result->values[string("b1")].at(1) < th) fitness++;
	if(result->values[string("b0")].at(1) > th) fitness++;
	if(result->values[string("c1")].at(1) < th) fitness++;
	if(result->values[string("c0")].at(1) > th) fitness++;
	if(result->values[string("d1")].at(1) > th) fitness++;
	if(result->values[string("d0")].at(1) < th) fitness++;


	delete result;
	result=NULL;

	// Case 4: a1=0,a0=1,b1=0,b0=1,c1=0,c0=1
	// set concentrations of inflow parameters 
	m->getParameter("inflowa1")->setValue(1.0);
 	m->getParameter("inflowa0")->setValue(0.0);
 	m->getParameter("inflowb1")->setValue(1.0);
 	m->getParameter("inflowb0")->setValue(0.0);
 	m->getSpecies("c1")->setInitialConcentration(0.0);
 	m->getSpecies("c0")->setInitialConcentration(1.0);
 	m->getSpecies("d1")->setInitialConcentration(1.0);
 	m->getSpecies("d0")->setInitialConcentration(0.0);


	// run simulation
	result = integrator->integrateSOSlib(m, time, printstep, false);

	if(result == NULL){
		return VERY_BAD_FITNESS;
	}

	// evaluate
	if(result->values[string("a1")].at(1) < th) fitness++;
	if(result->values[string("a0")].at(1) > th) fitness++;
	if(result->values[string("b1")].at(1) < th) fitness++;
	if(result->values[string("b0")].at(1) > th) fitness++;
	if(result->values[string("c1")].at(1) > th) fitness++;
	if(result->values[string("c0")].at(1) < th) fitness++;
	if(result->values[string("d1")].at(1) < th) fitness++;
	if(result->values[string("d0")].at(1) > th) fitness++;


	delete result;
	result=NULL;

	// Case 5: a1=0,a0=1,b1=1,b0=0,c1=1,c0=0
	// set concentrations of inflow parameters 
	m->getParameter("inflowa1")->setValue(1.0);
 	m->getParameter("inflowa0")->setValue(0.0);
 	m->getParameter("inflowb1")->setValue(0.0);
 	m->getParameter("inflowb0")->setValue(1.0);
 	m->getSpecies("c1")->setInitialConcentration(1.0);
 	m->getSpecies("c0")->setInitialConcentration(0.0);
 	m->getSpecies("d1")->setInitialConcentration(0.0);
 	m->getSpecies("d0")->setInitialConcentration(1.0);

	// run simulation
	result = integrator->integrateSOSlib(m, time, printstep, false);

	if(result == NULL){
		return VERY_BAD_FITNESS;
	}

	// evaluate
	if(result->values[string("a1")].at(1) < th) fitness++;
	if(result->values[string("a0")].at(1) > th) fitness++;
	if(result->values[string("b1")].at(1) > th) fitness++;
	if(result->values[string("b0")].at(1) < th) fitness++;
	if(result->values[string("c1")].at(1) > th) fitness++;
	if(result->values[string("c0")].at(1) < th) fitness++;
	if(result->values[string("d1")].at(1) < th) fitness++;
	if(result->values[string("d0")].at(1) > th) fitness++;


	delete result;
	result=NULL;

	// Case 6: a1=0,a0=1,b1=1,b0=0,c1=0,c0=1
	// set concentrations of inflow parameters 
	m->getParameter("inflowa1")->setValue(1.0);
 	m->getParameter("inflowa0")->setValue(0.0);
 	m->getParameter("inflowb1")->setValue(0.0);
 	m->getParameter("inflowb0")->setValue(1.0);
 	m->getSpecies("c1")->setInitialConcentration(0.0);
 	m->getSpecies("c0")->setInitialConcentration(1.0);
 	m->getSpecies("d1")->setInitialConcentration(1.0);
 	m->getSpecies("d0")->setInitialConcentration(0.0);


	// run simulation
	result = integrator->integrateSOSlib(m, time, printstep, false);

	if(result == NULL){
		return VERY_BAD_FITNESS;
	}

	// evaluate
	if(result->values[string("a1")].at(1) < th) fitness++;
	if(result->values[string("a0")].at(1) > th) fitness++;
	if(result->values[string("b1")].at(1) > th) fitness++;
	if(result->values[string("b0")].at(1) < th) fitness++;
	if(result->values[string("c1")].at(1) > th) fitness++;
	if(result->values[string("c0")].at(1) < th) fitness++;
	if(result->values[string("d1")].at(1) < th) fitness++;
	if(result->values[string("d0")].at(1) > th) fitness++;




	delete result;
	result=NULL;

	return (double) fitness;
}

