
#include <istream>
#include <sstream>
#include <fstream>

#include "../integrator.h"


using namespace sbmlevolver;

namespace sbmlevolver{

class ObjectiveFunction : public ObjectiveFunctionBase{
public:
	/// Default construnctor, uses file "objectivefunction.txt"
	ObjectiveFunction(unsigned long int seed, const char* filename = NULL);
	
	/// Construct the object from a string
	ObjectiveFunction(string, unsigned long int seed);
	
	/// Destructor
	~ObjectiveFunction();
	
	/// Print to the given stream
	void print(ostream&);	
	
private:
	/// Get the fitness of a model. Always specify this function!
	double calculateRawFitness(Model*);

	/// the initial concentrations of the input species (for io-table)
	vector<vector<vector<double> > > initialInputConcentrations; // Cases, species, timestep
	
	/// the desired output concentrations (for io-table)
	vector<vector<vector<double> > > outputConcentrations; // Cases, species, timesteps
	
	/// Read input
	void readInput(istream*);
	
	/// Abort at steady state?
	bool abortAtSteadyState;
	
	/// Get the initial concentration of input species j in case i
	double getInitialInputConcentration(int, int, int);
	
	/// Set initial concentration of input species in given case.
	void setInitialInputConcentrations(Model*, int, int);
	
	Integrator* integrator;
   
};

} // namespace sbmlevolver

