//#include "../objectivefunction.h"
// Not needed, you can use anything from ObjectiveFunction

using namespace sbmlevolver;

/// Default construnctor, may use an input file
ObjectiveFunction::ObjectiveFunction(unsigned long int seed, const char* filename){
	inputSpecies.push_back(string("inputA"));
	inputSpecies.push_back(string("inputB"));
	//outputSpecies.push_back(string("output"));
	integrator = new Integrator();
	this->seed = seed;
	rng = new RandomNumberGenerator(seed);
	
	nonMassConservancePenalty = 0;
	useAIC = false;
	sizePenalty = 0;
}
	
/// Construct the object from a string
ObjectiveFunction::ObjectiveFunction(string, unsigned long int seed){
	inputSpecies.push_back(string("inputA"));
	inputSpecies.push_back(string("inputB"));
	//outputSpecies.push_back(string("output"));
	integrator = new Integrator();
	this->seed = seed;
	rng = new RandomNumberGenerator(seed);
	
	nonMassConservancePenalty = 0;
	useAIC = false;
	sizePenalty = 0;
}
	
/// Destructor
ObjectiveFunction::~ObjectiveFunction(){
	delete integrator;
	delete rng;
}

/// Get the fitness of a model. Always specify this function!
double ObjectiveFunction::calculateRawFitness(Model* m){

	
	makeReactionsIrreversible(m);
	// SaveSBMLModel(m, "test2.xml");

	
	integrator_result* result;
	double fitness = 0;
	int init = 0;
	int step = 5;
	int num_replicates = 50;
	// Test off-behaviour up to 10
//cout << fitness << endl;
	for (init = 5; init < (2*step+0.1); init += step){
		cout << "GeneA: " << m->getSpecies("GeneA")->getInitialAmount() << endl;
		//		m->getSpecies("inputA")->setInitialAmount(init);
		//		m->getSpecies("inputB")->setInitialAmount(init);
		for(int i=0; i < num_replicates; i++){
SaveSBMLModel(m, "test.xml");
			result = integrator->integrateGillespie(m, 300, 300, rng->uniform(0,1000000000));
			if (result == NULL) return VERY_BAD_FITNESS;
			fitness += getStatusGeneA(result);
			fitness += getStatusGeneB(result);
			delete result;
		}
	}
//cout << fitness << endl;
	// Test on-behaviour down to 30
	for (init = 6*step; init < (8*step+0.1); init += step){
		m->getSpecies("inputA")->setInitialAmount(init);
		m->getSpecies("inputB")->setInitialAmount(init);
		for(int i=0; i < num_replicates; i++){
			result = integrator->integrateGillespie(m, 300, 300, rng->uniform(0,1000000000));
			if (result == NULL) return VERY_BAD_FITNESS;
			fitness += (1 - getStatusGeneA(result));
			fitness += (1 - getStatusGeneB(result));
			delete result;
		}
	}
//cout << fitness << endl;
	// Test switching behaviour
	int num_off = 0;
	int num_on = 0;
	for (init = 3*step; init < (5*step+0.1); init += step){
		m->getSpecies("inputA")->setInitialAmount(init);
		m->getSpecies("inputB")->setInitialAmount(init);
		for(int i=0; i < num_replicates; i++){
			result = integrator->integrateGillespie(m, 300, 300, rng->uniform(0,1000000000));
			if (result == NULL) return VERY_BAD_FITNESS;
			if(getStatusGeneA(result)) num_on++;	
			else num_off++;
			if(getStatusGeneB(result)) num_on++;	
			else num_off++;
			delete result;
		}
		if(num_on > num_off) fitness += (num_off/num_replicates);
		else fitness += (num_on/num_replicates);
		num_on = 0;
		num_off = 0;
	}
//cout << fitness << endl;
//cout << endl;
	fitness = fitness / num_replicates;
	return fitness;
}

// Gene is "on" if it is activated at any of the last 100 steps
int ObjectiveFunction::getStatusGeneA(integrator_result* res){
	int timepoints = res->timepoints.size();
	if(timepoints < 100) THROW_ERROR("Not enough timepoints in results object!");
	int counter = 0;
	for(int i=1; i <= 99; i++){
		if(res->values["GeneAAct"].at(timepoints-i) == 1) counter++;
	}
	if(counter > 0) return 1; // Gene on
	else return 0; // Gene off
}
// Gene is "on" if it is activated at any of the last 100 steps
int ObjectiveFunction::getStatusGeneB(integrator_result* res){
	int timepoints = res->timepoints.size();
	if(timepoints < 100) THROW_ERROR("Not enough timepoints in results object!");
	int counter = 0;
	for(int i=1; i <= 99; i++){
		if(res->values["GeneBAct"].at(timepoints-i) == 1) counter++;
	}
	if(counter > 0) return 1;
	else return 0;
}

