/* Version information
 *
 * $Id: paramobj.cpp 29 2007-06-15 13:35:45Z thlenser $
 *
 */


#include "paramobj.h"

using namespace std;
using namespace sbmlevolver;

ParamObj::ParamObj(){
	// init the standard values
	initDefaults();
}

ParamObj::ParamObj(string optionsfile) {
	// init the standart values
	initDefaults();
	
	// Parse the options file
	parseOptionsFile(optionsfile);
}

	
ParamObj::~ParamObj() {
	// free vectors
	mutator.mutationProbs.clear();
}



/**
 * Parse the options file. This file has format 'parameter	value'
 * on each line
 */
#define IGNORE_WHITESPACE while((input->peek() == ' ') || (input->peek() == '\t')) input->ignore();
#define NEXT_LINE if(input->peek() == '\n') input->ignore();
		  //else THROW_ERROR("Unexpected data while reading options");
string findWord(istream* input){
	string tmp_str;
	char* tmp_char = new char();
	while((input->peek() != ' ') && (input->peek() != '\t') && (input->peek() != '\n')){
		input->get(tmp_char,2);
		tmp_str.append(tmp_char);
	}
	delete tmp_char;
	return tmp_str;
}

string nextWord(istream* input) {
	char* c=new char();
	string result = "";
	if ( input->eof() ) throw "can't read word, end of input reached!";
	while((input->peek() == ' ') || (input->peek() == '\t') || (input->peek() == '\n')) input->ignore();
	while ( input->good() ) {
		input->get(c,2,NULL );
		switch ( *c ) {
			case '\n':
			case '\t':
			case ' ':
        delete c;
        return result;
				break;
		}
		result.append(c);
	}
	delete c;
	return result;
}

void ParamObj::parseOptionsFile(string filename) {
	ifstream* input = new ifstream();
	input->open(filename.c_str());
	if(input->fail()){
		cout << "Could not find " << filename << endl;
		THROW_ERROR("No optionsfile with given filename exists!");
	}
	
	parseOptions((istream*) input);
	
	input->close();
	delete input;
}

int nextInt(istream* input) {
	int result;
	if ((*input >> result).fail()) throw "Expected integer value, bot none found!";
	return result;
}

double nextDouble(istream* input) {
	double result;
	if ((*input >> result).fail()) throw "Expected double value, bot none found!";
	return result;
}

long nextLong(istream* input) {
	long result;
	if ((*input >> result).fail()) throw "Expected longint value, bot none found!";
	return result;
}

string readFromFile(string filename) {
	string result;
  char c;
	// Try to open file
	ifstream f_input;
	f_input.open(filename.data());
	if ( f_input ) { // the file is there
		while ( !f_input.eof() ) {
			f_input.get(c);
			result.push_back(c);
		}
	} else {
		throw "Cannot find specified file";
	}
	f_input.close();
	return result;
}

void ParamObj::parseOptions(istream* input) {

	string word;
	char c;
	bool switch_known;
	while ( !input->eof() ) {
		while ( (input->peek() == '#') || (input->peek() == '\n') )	input->ignore(10000, '\n');

		if ( input->eof() ) break;
		word = nextWord(input);
		//cout << word << endl;
		switch_known = false;
		if ( word == "mu" && ++switch_known ) population.mu = nextInt(input);
		if ( word == "lambda" && ++switch_known ) population.lambda = nextInt(input);
		if ( word == "num_pop" && ++switch_known ) population.num_populations = nextInt(input);
		if ( word == "iso_time" && ++switch_known ) population.isolationTime = nextInt(input);
		if ( word == "first_model" && ++switch_known ) population.firstModelFile = nextWord(input);
		if ( word == "init_num_species" && ++switch_known ) population.initialNumSpecies = nextInt(input);
		if ( word == "init_num_reactions" && ++switch_known ) population.initialNumReactions = nextInt(input);
		if ( word == "cross_prob" && ++switch_known ) population.crossover_prob = nextDouble(input);
		if ( word == "comma_selection" && ++switch_known ) population.comma = true;
		if ( word == "always_evaluate" && ++switch_known ) population.alwaysEvaluate = true;
		if ( word == "output_file" && ++switch_known ) session.outputFile = nextWord(input);
		//  if( word == "input_dir" && ++switch_known )					session.inputDir = nextWord(input);
		if ( word == "output_dir" && ++switch_known ) session.outputDir = nextWord(input);
		if ( word == "bblocks" && ++switch_known ) {
			string filename=nextWord(input);
			try {
			  BBstring = readFromFile(filename);
			} catch (char* e){
				cerr << e << filename << endl;
				THROW_ERROR("Cannot find specified building-blocks file");
			}
		}
		if ( word == "objective" && ++switch_known ) {
			string filename=nextWord(input);
			try {
				OFstring = readFromFile(filename);
			} catch (char* e){
				cerr << e << filename << endl;
				THROW_ERROR("Cannot find specified objective-function file");
			}
		}
		if ( word == "save_every" && ++switch_known ) session.saveEvery = nextInt(input);
		if ( word == "num_save" && ++switch_known ) session.numSave = nextInt(input);
		if ( word == "stop_fitness" && ++switch_known ) session.stopFit = nextDouble(input);
		if ( word == "stop_turn" && ++switch_known ) session.stopTurn = nextLong(input);
		if ( word == "stop_time" && ++switch_known ) *input >> session.stopTime;
		if ( word == "mutation_probs" && ++switch_known ) {
			mutator.mutationProbs.clear();
			for ( int i = 0; i < 7; i++ ) mutator.mutationProbs.push_back(nextDouble(input));
		}
		if ( word == "max_mutations" && ++switch_known ) mutator.maxSimultan = nextInt(input);
		if ( word == string("pf_setting") && ++switch_known ) {
			paraFitter.numOfParents = nextInt(input);
			paraFitter.populationSize = nextInt(input);
			if(paraFitter.populationSize < 2){
				cout << "Warning: parameter fitter population size cannot be <2, set to 2" << endl;
				paraFitter.populationSize = 2;
			}
			paraFitter.numOfGenerations = nextInt(input);
		}
		if ( word == "pf_no_inheritance" && ++switch_known ) paraFitter.useOldStartingValues = false;
		if ( word == "seed" && ++switch_known ) seed = nextLong(input);
		if ( !switch_known ) {
			cout << "Encountered unknown argument " << word << " while reading options" << endl;
			if (input->bad()) cout << "input went bad" << endl;
			if (input->fail()) cout << "input fail" << endl;

			exit(-1);
		}
	}
}

/*void ParamObj::parseOptions(istream* input) {
 string tmp_str;
 char tmp_char;
 while ( !input->eof() ) {
 while ( (input->peek() == '#') || (input->peek() == '\n') )
 input->ignore(10000, '\n');
 if ( input->eof() ) break;
 tmp_str.clear();
 tmp_str = findWord(input);
 if ( tmp_str == string( "mu" ) ) *input >> population.mu;
 else if ( tmp_str == string( "lambda" ) ) *input >> population.lambda;
 else if ( tmp_str == string( "num_pop" ) ) *input >> population.num_populations;
 else if ( tmp_str == string( "iso_time" ) ) *input >> population.isolationTime;
 else if ( tmp_str == string("first_model") ) {
 IGNORE_WHITESPACE
 population.firstModelFile = findWord(input);
 IGNORE_WHITESPACE
 } else if ( tmp_str == string( "init_num_species" ) ) *input >> population.initialNumSpecies;
 else if ( tmp_str == string( "init_num_reactions" ) ) *input >> population.initialNumReactions;
 else if ( tmp_str == string( "cross_prob" ) ) *input >> population.crossover_prob;
 else if ( tmp_str == string( "comma_selection") ) population.comma = true;
 else if ( tmp_str == string( "always_evaluate") ) population.alwaysEvaluate = true;
 else if ( tmp_str == string("output_file") ) {
 IGNORE_WHITESPACE
 session.outputFile = findWord(input);
 IGNORE_WHITESPACE
 }
 /*else if(tmp_str == string("input_dir")) {
 IGNORE_WHITESPACE
 session.inputDir = findWord(input);
 IGNORE_WHITESPACE
 }/
 else if ( tmp_str == string("output_dir") ) {
 IGNORE_WHITESPACE
 session.outputDir = findWord(input);
 IGNORE_WHITESPACE
 } else if ( tmp_str == string("bblocks") ) {
 BBstring.clear();
 IGNORE_WHITESPACE
 // Try to open building blocks file
 ifstream bbinput;
 tmp_str = findWord(input);
 bbinput.open(tmp_str.c_str());
 if ( bbinput ) { // the file is there
 while ( !bbinput.eof() ) {
 bbinput.get(tmp_char);
 BBstring.push_back(tmp_char);
 }
 } else {
 cerr << tmp_str << endl;
 THROW_ERROR("Cannot find specified building-blocks file");
 }
 bbinput.close();
 IGNORE_WHITESPACE
 } else if ( tmp_str == string("objective") ) {
 OFstring.clear();
 IGNORE_WHITESPACE
 tmp_str = findWord(input);
 // Try to open objective function file
 ifstream ofinput;
 ofinput.open(tmp_str.c_str());
 if ( ofinput ) { // the file is there
 while ( !ofinput.eof() ) {
 ofinput.get(tmp_char);
 OFstring.push_back(tmp_char);
 }
 } else {
 cerr << tmp_str << endl;
 THROW_ERROR("Cannot find specified objective-function file");
 }
 ofinput.close();
 IGNORE_WHITESPACE
 } else if ( tmp_str == string( "save_every" ) ) *input >> session.saveEvery;
 else if ( tmp_str == string( "num_save" ) ) *input >> session.numSave;
 else if ( tmp_str == string( "stop_fitness" ) ) *input >> session.stopFit;
 else if ( tmp_str == string( "stop_turn" ) ) *input >> session.stopTurn;
 else if ( tmp_str == string( "stop_time" ) ) *input >> session.stopTime;
 else if ( tmp_str == string("mutation_probs") ) {
 mutator.mutationProbs.clear();
 double tmp;
 for ( int i = 0; i < 7; i++ ) {
 *input >> tmp;
 mutator.mutationProbs.push_back(tmp);
 }
 } else if ( tmp_str == string( "max_mutations" ) ) *input >> mutator.maxSimultan;
 else if ( tmp_str == string("pf_setting") ) {
 *input >> paraFitter.numOfParents;
 *input >> paraFitter.populationSize;
 *input >> paraFitter.numOfGenerations;
 } else if ( tmp_str == string("pf_no_inheritance") ) paraFitter.useOldStartingValues = false;
 else if ( tmp_str == string( "seed" ) ) *input >> seed;
 else {
 cout << "Encountered unknown argument " << tmp_str << " while reading options" << endl;
 exit(-1);
 }
 IGNORE_WHITESPACE
 NEXT_LINE
 }
 }*/

void ParamObj::initDefaults() {
	int i;
	char tmp_char;
	ifstream input;
	ifstream input2;

	// seed
	char hostname[255];
	gethostname(hostname,255);
	seed = 0;
	for(int i=0; (i < 255) && (hostname[i] != '\0'); i++){
		seed += (int) hostname[i];
	}
	seed += (unsigned long int) time(NULL) + getpid();
	
	/* params */
	session.stopFit = 0.0;
	session.stopTurn = 0;
	session.stopTime = 0;
	
	session.saveEvery = 10;
	session.numSave = 0;
	
	/* Try to read OF and BB from default files */
	if(BBstring.size() == 0){
		input.open("buildingblocks.txt");
		if(input){	// the file is there
			while(! input.eof()){
				input.get(tmp_char);
				BBstring.push_back(tmp_char);
			}
		}
		input.close();
	}
							
	if(OFstring.size() == 0){
		input2.open("objectivefunction.txt");
		if(input2){	// the file is there
			while(! input2.eof()){
				input2.get(tmp_char);
				OFstring.push_back(tmp_char);
			}
		}
		input2.close();
	}
	
	population.num_populations = 1;
	population.isolationTime = 0;
	population.mu = 10;
	population.lambda = 100;
	population.crossover_prob = 0.5;
	population.comma = false;
	population.alwaysEvaluate = false;
	population.initialNumSpecies = 0;
	population.initialNumReactions = 0;
	
	for (i=0;i<7;i++) mutator.mutationProbs.push_back(1);
	mutator.maxSimultan = 1;
	
	paraFitter.populationSize = 20;
	paraFitter.numOfParents = 10;
	paraFitter.numOfGenerations = 50;
	paraFitter.useOldStartingValues = true;
}



/**
 * Write the whole data stored in this object to a ostream.
 * This should be called from centralcontrol in order to store the session parameters.
 */
void ParamObj::print(ostream &outstream){
	outstream << "mu " << population.mu << endl;
	outstream << "lambda " << population.lambda << endl;
	outstream << "num_pop " << population.num_populations << endl;
	outstream << "iso_time " << population.isolationTime << endl;
	outstream << "first_model " << population.firstModelFile << endl;
	outstream << "initial num of species " << population.initialNumSpecies << endl;
	outstream << "initial num of reactions " << population.initialNumReactions << endl;
	outstream << "cross_prob " << population.crossover_prob << endl;
	if(population.comma) outstream << "comma_selection " << endl;
	if(population.alwaysEvaluate) outstream << "always_evaluate " << endl;
	outstream << "output_file " << session.outputFile << endl;
	//outstream << "input_dir " << session.inputDir << endl;
	outstream << "output_dir " << session.outputDir << endl;
	outstream << "save_every " << session.saveEvery << endl;
	outstream << "num_save " << session.numSave << endl;
	outstream << "stop_fitness " << session.stopFit << endl;;
	outstream << "stop_turn " << session.stopTurn << endl;
	outstream << "stop_time " << session.stopTime << endl;
	outstream << "mutation_probs";
	for(int i=0; i < mutator.mutationProbs.size(); i++){
		outstream << " " << mutator.mutationProbs.at(i);
	}
	outstream << endl;
	outstream << "max_mutations " << mutator.maxSimultan << endl;
	outstream << "pf_mu " << paraFitter.numOfParents << endl;
	outstream << "pf_lambda " << paraFitter.populationSize << endl;
	outstream << "pf_num_gen " << paraFitter.numOfGenerations << endl;
	outstream << "seed " << seed << endl;
	outstream << "bblocks " << BBstring << endl;
	outstream << "objective " << OFstring << endl;
}

/**
 * Describe all parameter in text form
 * Can be called from centralcontrol as a help-facility
 */
void ParamObj::printHelp(ostream &outstream){
	// Boolean parameters are marked with '+'
	// Parameter groups are marked with '*'
	outstream << "# number of populations" << endl;
	outstream << "num_pop [int]" << endl;
	outstream << "# time between exchanges between populations" << endl;
	outstream << "iso_time [int]"  << endl;
	outstream << "# population size" << endl;
	outstream << "mu [int]" << endl;
	outstream << "# number of offspring" << endl;
	outstream << "lambda [int]" << endl;
	outstream << "# width of fitness-sharing kernel" << endl;
	outstream << "# The initial solution to start from" << endl;
	outstream << "first_model [string]" << endl;
	outstream << "# Average number of species in initial population" << endl;
	outstream << "init_num_species [int]" << endl;
	outstream << "# Average number of reactions in initial population" << endl;
	outstream << "init_num_reactions [int]" << endl;
	outstream << "# Probability for crossover" << endl;
	outstream << "cross_prob [double]" << endl;
	outstream << "# Comma selection instead of default plus selection" << endl;
	outstream << "+comma_selection " << endl;
	outstream << "# Re-evaluate population in each generation (for stochastic fitness function)" << endl;
	outstream << "+always_evaluate " << endl;
	outstream << "# Copy output into this file" << endl;
	outstream << "output_file [string]" << endl;
	//outstream << "# Read input from this directory" << endl;
	//outstream << "input_dir [string]" << endl;
	outstream << "# Write output into this directory" << endl;
	outstream << "output_dir [string]" << endl;
	outstream << "# Number of generations between saving of individuals" << endl;
	outstream << "save_every [int]" << endl;
	outstream << "# Number of individuals to save every 'save_every' generation" << endl;
	outstream << "num_save [int]" << endl;
	outstream << "# Stop evolution when this fitness is reached" << endl;
	outstream << "stop_fitness [double]" << endl;;
	outstream << "# Stop evolution when this generation is reached" << endl;
	outstream << "stop_turn [int]"  << endl;
	outstream << "# Stop evolution when this time is reached (in seconds)" << endl;
	outstream << "stop_time [int]"  << endl;
	outstream << "# Relative mutation probabilities (no mutation, add/delete species, replace reaction, add/delete reaction, duplicate species)" << endl;
	outstream << "*mutation_probs (7 numbers) [double]" << endl;
	outstream << "# Maximal number of mutations per turn per individuals" << endl;
	outstream << "max_mutations [int]" << endl;
	outstream << "# Population size, number of offspring and number of generations for parameter fitter" << endl;
	outstream << "*pf_setting (3 numbers) [double]" << endl;
	outstream << "# Building-blocks file" << endl;
	outstream << "bblocks [string]" << endl;
	outstream << "# Objective-function file" << endl;
	outstream << "objective [string]" << endl;
}
