/* Version information
 *
 * $Id: population.cpp 3 2007-06-04 12:59:46Z thlenser $
 *
 */


#include "population.h"

using namespace std;
using namespace sbmlevolver;


Population::Population(unsigned long int seed, ParamObj* params) {
	// init
	HighestId = 0;
	CurrentGeneration = 0;
	mu = params->population.mu;
	lambda = params->population.lambda;
	comma = params->population.comma;
	crossover_prob = params->population.crossover_prob;

	// init main sub objects
	rng = new RandomNumberGenerator(seed);
	of = new ObjectiveFunction(params->OFstring, rng->getASeed());
	bb = new BuildingBlocks(params->BBstring);
	//paraFit = new ParameterFitter(params->paraFitter.populationSize, params->paraFitter.numOfParents, params->paraFitter.numOfGenerations, rng->getASeed());
	mut = new Mutator(rng->getASeed(), bb, params->mutator.mutationProbs, params->mutator.maxSimultan);
	
	// load first individual from file
	if (!(params->population.firstModelFile.empty())) {
		FirstModel = new Individual(params->population.firstModelFile.c_str(), 0, 0, of);
		
		// learn any new reaction types that are in the model
		// params->getBuildingBlocks()->learnReactionsFromModel(FirstModel->getModelPtr());
		// Disabled for the moment
	}else{
		FirstModel = new Individual(0,0,of,bb);
	}
	
	//FirstModel->calculateFitness(paraFit,of);
	
	// initialise population
	DIR *dir_handle; // directory entry
	dirent *dir;
	
	if ((dir_handle = opendir(".")) == NULL) THROW_ERROR("could not open result directory (permissions?)");
	while ((dir = readdir(dir_handle)) != NULL) {
		// check if filename ends with ".model.xml"
		if (((strstr((char*) &(dir->d_name), ".model.xml") - (char*) &(dir->d_name)) != (strlen((char*) &(dir->d_name)) - strlen(".model.xml")))) continue;
		
		Individual* ind = new Individual((char*) &(dir->d_name), ++HighestId, CurrentGeneration, of);
		openTasks.push_back(ind);

		cout << "added " << (char*) &(dir->d_name) << " to the population" << endl;
	}
	closedir(dir_handle); 
	// create further initial offspring until pop'size is reached
	while (openTasks.size() < mu) {
		// Create a new model from FirstModel
		Individual* ind = new Individual(FirstModel->getModelCopy(), ++HighestId, CurrentGeneration);
		
		//ind->mutateModel();
		mut->createRandomModel(ind,params->population.initialNumSpecies,params->population.initialNumReactions);
		
		// Add it to the population
		openTasks.push_back(ind);
	}
	
	openTasksOutsidePop = 0;
}

Population::~Population() {
	multimap<double, Individual*>::iterator iter;
	
	// delete all individuals
	iter = inds.begin();
	while (iter != inds.end()) {
		delete iter->second;
		iter++;
	}
	
	// free main sub objects
	delete FirstModel;
	delete rng; // free random number generator
	delete of;
	delete bb;
	delete mut;
}



void Population::unsetFitnessOfWholePopulation(){
	multimap<double, Individual*>::iterator iter;
	
	for (iter = inds.begin(); iter != inds.end(); iter++) {
		iter->second->unsetFitness();
		openTasks.push_front(iter->second);
	}
	inds.clear();
}
		

/*void Population::fitnessSharing(){
	if(alpha <= 0) return;
	if (inds.size() <= 0) THROW_ERROR("fitnessSharing() called, but no rated individuals are available");
	double f = alpha;
	if(inds.size() != 1){
		multimap<double,Individual*>::iterator ind;
		multimap<double,Individual*>::iterator ind2;
		for(ind = inds.begin(); ind != inds.end(); ind++){
			f = 0;
			for(ind2 = inds.begin(); ind2 != inds.end(); ind2++){
				if(ind != ind2){
					//f += exp(-alpha * ind->second->structuralDifference2(bb, ind2->second));
					f += exp(- pow(ind->second->structuralDifference2(bb, ind2->second),2.0) / pow(alpha,2.0));
				}
				//f *= alpha;
				//f /= 5.1132*alpha*inds.size(); // 2*sqrt(2*pi)*alpha, since it has only positive arguments
				f /= 5.1132*alpha; // 2*sqrt(2*pi)*alpha, since it has only positive arguments
			}
			if(!(ind->second->isFitness())) THROW_ERROR("fitnessSharing cannot be called on unrated individuals!");
			//cout << (1+f) * ind->second->getRawFitness() << endl;
			ind->second->setFitness((1.0+f) * ind->second->getRawFitness());
			if(ind->second->getFitness() > VERY_BAD_FITNESS) ind->second->setFitness(VERY_BAD_FITNESS);
		}
	}

	// Re-sort multimaps with new fitnesses
	multimap<double,Individual*> tmpMM;
	multimap<double,Individual*>::iterator ind;
	for(ind = inds.begin(); ind != inds.end(); ind++){
		tmpMM.insert(pair<double,Individual*>(ind->second->getFitness(),ind->second));
	}
	inds.swap(tmpMM);
	tmpMM.clear();

	// Now delete species tags
	//for(ind = inds.begin(); ind != inds.end(); ind++){
		//ind->second->clearTags();
		//}
}*/

    
void Population::incCurrentGeneration() {
	CurrentGeneration++;
}


IndGen Population::getCurrentGeneration() {
	return CurrentGeneration;
}

    
/*** population & selection parameters ***/

int Population::getMu() {
	return mu;
}

int Population::getLambda() {
	return lambda;
}
    
/*** population stats & managment ***/

/* do _not_ call this while the EP is within the poplation object
   these pointer invalidate immideatly after restarting population ! */
set<Individual*> *Population::getBestIndividuals(int numSave) {
	set<Individual*> *s;
	multimap<double, Individual*>::iterator iter;
	
	s = new set<Individual*>;
	
	iter = inds.begin();
	int counter = 0;
	while ((counter < numSave) && (iter != inds.end())) {
		s->insert(iter->second);
		iter++;
		counter++;
	}
	
	return s;
}


/*double Population::getBestRawFitness(){
	double best = VERY_BAD_FITNESS;
	multimap<double,Individual*>::iterator ind = inds.begin();
	for(; ind != inds.end(); ind++){
		if(ind->second->getRawFitness() < best) best = ind->second->getRawFitness();
	}
	return best;
}

double Population::getAverageRawFitness(){
	if (inds.size() > 0) {
		double avg = 0;
		multimap<double,Individual*>::iterator ind = inds.begin();
		for(; ind != inds.end(); ind++){
			avg += ind->second->getRawFitness();
		}
		return avg/inds.size();
	}else{
		THROW_ERROR("getAveragetFitness() called, but no fitness is available");
	}
}*/


double Population::getBestFitness() {
	if (inds.size() > 0) {
		return inds.begin()->first;
	}else{
		THROW_ERROR("getBestFitness() called, but no fitness is avaible");
	}
}

double Population::getWorstFitness() {
	if (inds.size() > 0) {
		return inds.rbegin()->first;
	}else{
		THROW_ERROR("getWorstFitness() called, but no fitness is avaible");
	}
}

double Population::getAverageFitness() {
	if (inds.size() > 0) {
		vector<double> fit = getAllFitnesses();
		double avg = 0;
		for(int i=0; i < fit.size(); i++) avg += fit.at(i);
		return avg / inds.size();
	}else{
		THROW_ERROR("getAveragetFitness() called, but no fitness is available");
	}
}

double Population::getVariance(){
	double var = 0;
	if (inds.size() > 0) {
		if(inds.size() == 1) var = 0;
		else{
			multimap<double,Individual*>::iterator ind = inds.begin();
			for(; ind != inds.end(); ind++){
				var += pow(ind->first - getAverageFitness(), 2);
			}
			var /= (inds.size()-1);
		}
		return var;
	}else{
		THROW_ERROR("getVariance() called, but no fitness is available");
	}
}


double Population::getAverageStructuralDifference(){
	double diff = 0;
	if (inds.size() <= 0) THROW_ERROR("getAverageStructuralDifference() called, but no rated individuals are available");
	if(inds.size() == 1) diff = 0;
	else{
		multimap<double,Individual*>::iterator ind;
		multimap<double,Individual*>::iterator ind2;
		for(ind = inds.begin(); ind != inds.end(); ind++){
			for(ind2 = inds.begin(); ind2 != ind; ind2++){
				diff += ind->second->structuralDifference(ind2->second);
			}
		}
		diff /= (inds.size() * inds.size() / 2.0);
	}
	return diff;
}

double Population::getAverageStructuralDifference2(){
	double diff = 0;
	if (inds.size() <= 0) THROW_ERROR("getAverageStructuralDifference() called, but no rated individuals are available");
	if(inds.size() == 1) diff = 0;
	else{
		multimap<double,Individual*>::iterator ind;
		multimap<double,Individual*>::iterator ind2;
		for(ind = inds.begin(); ind != inds.end(); ind++){
			for(ind2 = inds.begin(); ind2 != ind; ind2++){
				diff += ind->second->structuralDifference2(bb, ind2->second);
			}
		}
		diff /= (inds.size() * inds.size() / 2.0);
	}
	
	// Now delete species tags
	multimap<double,Individual*>::iterator ind;
	for(ind = inds.begin(); ind != inds.end(); ind++){
		ind->second->clearTags();
	}
	
	return diff;
}


double Population::getAverageNumSpecies(){
	double tmp = 0;
	if (inds.size() <= 0) THROW_ERROR("getAverageNumSpecies() called, but no rated individuals are available");
	multimap<double,Individual*>::iterator ind;
	for(ind = inds.begin(); ind != inds.end(); ind++){
		tmp += ind->second->getModelPtr()->getNumSpecies();
	}
	tmp /= (double) inds.size();
	return tmp;
}

double Population::getAverageNumReactions(){
	double tmp = 0;
	if (inds.size() <= 0) THROW_ERROR("getAverageNumReactions() called, but no rated individuals are available");
	multimap<double,Individual*>::iterator ind;
	for(ind = inds.begin(); ind != inds.end(); ind++){
		tmp += ind->second->getModelPtr()->getNumReactions();
	}
	tmp /= (double) inds.size();
	return tmp;
}
	


vector<double> Population::getAllFitnesses(){
	vector<double> fit;
	fit.reserve(inds.size());
	multimap<double,Individual*>::iterator ind = inds.begin();
	for(; ind != inds.end(); ind++){
		fit.push_back(ind->first);
	}
	return fit;
}


/**
 * Cut the number of individuals down to mu
 * If it is lower than mu, throw an error
 */
void Population::select(){
	if(inds.size() < mu) THROW_ERROR("Not enough individuals to select from!");
	if (openTasks.size() > 0) THROW_ERROR("Not all individuals have a fitness when selecting!");
	while (inds.size() > mu){
		if (inds.size() > 0) removeIndividual(inds.rbegin()->second);
		else THROW_ERROR("Something is wrong in the selection scheme!");
	}
}


/**
 * Create lambda offspring
 * default is "+" selection, i.e. offspring are added to population.
 * if comma=TRUE, the old population is discarded
 */
void Population::createOffspring(){
	if (openTasks.size() > 0){
		if(CurrentGeneration == 1) return;
		else THROW_ERROR("Not all individuals have a fitness when creating offspring!");
	}

	// create new individuals
	if (inds.size() == 0) {
		// we have to create new ind's from scratch
		for(int i=0; i < lambda; i++){
			// NOT: addIndividual(FirstModel);
			// hardcopy FirstModel first !
			Individual* tmp =  new Individual(FirstModel->getModelCopy(),++HighestId,CurrentGeneration);
			openTasks.push_back(tmp);
		}
	}else{
		// we create offspring from Ind's with known fitness
		for(int i=0; i < lambda; i++){
			Individual* tmp;
			if(rng->challenge(crossover_prob)) tmp = mut->recombine(RandomSelectInd(),RandomSelectInd(), ++HighestId,CurrentGeneration);
			else tmp = mut->getMutant(RandomSelectInd(), ++HighestId, CurrentGeneration);
			openTasks.push_back(tmp);
		}
	}
	
	if(comma){
		// delete old population
		while(inds.size() > 0){
			removeIndividual(inds.rbegin()->second);
		}
	}
}
	

    
/*** individuals ***/
void Population::addIndividual (Individual* ind) {
	// Check if individual has fitness
	if (ind->isFitness()) {
		inds.insert(inds.begin(), pair<double, Individual*>(ind->getFitness(), ind));
	}else{
		THROW_ERROR("Trying to add an individual with unknown fitness to the population!");
	}
}

// received ind from other population
void Population::receiveIndividual(Individual* ind){
	if(ind != NULL){
		ind->setId(++HighestId);
		ind->setGeneration(CurrentGeneration);

		addIndividual(ind);
	}
}

// extract (a copy of) the best individual, which can then migrate
Individual* Population::extractBestIndividual(){
	//if(inds.size() > 0) return new Individual(inds.begin()->second->getModelCopy(), 0, 0);
	if(inds.size() > 0) return new Individual(inds.begin()->second);
	else THROW_ERROR("No individual with fitness is available!");
}

void Population::removeIndividual (Individual* ind) {
	// remove fitness specific stuff
	if (ind->isFitness()) {
		multimap<double, Individual*>::iterator iterFit;
		
		// 1. delete from fitness index
		iterFit = inds.find(ind->getFitness());
		
		while (iterFit != inds.end() && iterFit->second->getId() != ind->getId()) iterFit++;
		if (iterFit == inds.end()) THROW_ERROR("an individual to be removed was not found in inds");
	
		inds.erase(iterFit);
		
	}else{
		THROW_ERROR("Trying to remove an individual with unknown fitness!");
	}
	delete ind;
}



/* random selection

   METHOD
     * RandomSelectInd uses an equaly distributed selection
     * in RandomSelectInd_WeightUnfit each Ind has the propability
       (Fitness / SumOfFitness) to get choosen -> linear proportionality
     * RandomSelectInd_WeightFit not implemented yet

   HINTs
     * all selection functions select from the pool of Ind's w/ _known_ fitness
       even if RandomSelectInd does not rely on
     * they are also only thought to operate w/ all Ind's tested
       ([UnratedQueue] = 0) but should have no problem if this is not the case */

Individual *Population::RandomSelectInd() {
	multimap<double, Individual*>::iterator iter;
	int PartNum; // num of Ind's still avaible for random selection
	
	if (inds.size() == 0) THROW_ERROR("RandomSelectInd() called but no Ind's avaible");
	
	PartNum = inds.size();
	iter = inds.begin();
	
	while (iter != inds.end()) {
		if (rng->challenge(1 / ((double) PartNum))) return iter->second;
		
		PartNum--;
		iter++;
	}
	
	THROW_WARN("rng->challenge seems to be broken: event with propability 1 did not incide")
	
	iter--;
	return iter->second;
}

/*Individual *Population::RandomSelectInd_WeightFit() {
	return RandomSelectInd();
}*/

// Leave the best numEliteSurvivors alone...
Individual *Population::RandomSelectInd_WeightUnfit(int numEliteSurvivors) {
	multimap<double, Individual*>::reverse_iterator iter;
	double PartSum; // the partial sum: SumFitness without the fitness of already passed Ind's
	
	// Get sum of all individuals which can be selected for removal
	// Leave the best numEliteSurvivors alone
	if (inds.size() == 0) THROW_ERROR("RandomSelectInd() called but no Ind's avaible");
	vector<double> fit = getAllFitnesses();
	PartSum = 0;
	for(int i=numEliteSurvivors; i < fit.size(); i++) PartSum += fit.at(i);

	if (PartSum <= 0) THROW_ERROR("Fitness of elite survivors is greater than overall fitness sum.");

	// select randomly
	iter = inds.rbegin();
	while (iter != inds.rend()) {
		if (rng->challenge(iter->first / PartSum)) return iter->second;
		PartSum -= iter->first;
		iter++;
	}
	
	/* the last Ind in the loop below should have the propability 1.0 so
	   we're facing at least a precision problem here */

	THROW_ERROR("SumFitness is incorrect (should not happen in RandomSelectInd_WeightUnfit) or RandomNumberGenerator does not work");
		
	iter--;
	return iter->second;
}


Individual* Population::extractTask() {
	// get the individual
	if(openTasks.size() <= 0) return NULL;
	Individual* ind = openTasks.front();
	openTasks.pop_front();
		
	// log that our thread is working now
	openTasksOutsidePop++;
	
	return ind;
}

void Population::cancelTask(Individual *ind) {
	// re-add our previous task
	openTasks.push_back(ind);
	
	// our thread does not work anymore
	openTasksOutsidePop--;
	
	// may be other threads wait for work already ?
}

void Population::submitResult(Individual *ind) {
	// enter result
	addIndividual(ind);
	
	openTasksOutsidePop--;
}

int Population::tasksInProgress(){
	return openTasksOutsidePop;
}
