/************************************************************************
		sbmltools.h - SBMLEvolver - CSB, FSU Jena

This software is free for use and modification, but comes without any
warranty. Usage is entirely at your own risk!


These tools are for anything that is only related to SBML handling,
serving as an extension to the libSBML library. They are mainly
concerned with copying different parts of SBML objects, but some
other functionality is also offered.

	$Id: sbmltools.h,v 1.4 2007/06/11 09:04:27 thlenser Exp $

*************************************************************************/

#ifndef SBMLTOOLS_H_
#define SBMLTOOLS_H_

#include <vector>
#include <sstream>
#include <string.h>

#include "include.h"
#include "SBMLTypes.h"

using namespace std;

namespace sbmlevolver{

/// Replace species sp1 in reaction re with species sp2
void replaceSpeciesInReaction(Reaction* re, string sp1, string sp2);
/// Check if two reactions are equal
bool equal(Reaction*, Reaction*);

/// Replace all reversible reactions by two irreversible ones
void makeReactionsIrreversible(Model* mod);

/// make sure each species appears only one as reactant and product
void cleanStoichiometries(Model* m);
	

/// Write SBML model to a string
string SerializeSBMLModel(Model*);
/// Read SBML model from a string
Model* DeserializeSBMLModel(string);

/// Write SBML model to a file
Model* LoadSBMLModel(const char* file_name);
/// Read SBML model from a file
void SaveSBMLModel(Model* m, const char* file_name);

/// Lots of functions for manipulating the Annotation tag in SBase elements
bool isMutational(SBase* sb);
void setMutational(SBase* sb, bool mutational);
void insertIntoAnnotation(SBase* sb, string insertme);
void removeFromAnnotation(SBase* sb, string removeme);

void checkAttrName(string attr);
void setAttr(SBase* sb, string attr, double val);
void setAttr(SBase* sb, string attr, string val);
double getAttr(SBase* sb, string attr);
bool isSetAttr(SBase* sb, string attr);
void unsetAttr(SBase* sb, string attr);

string getCompartmentOfReaction(Model* m, Reaction* re);
vector<int> getSpeciesInCompartment(Model* m, string compartment);

void printReaction(Reaction* re);

}; // namespace sbmlevolver

#endif /*SBMLTOOLS_H_*/
